'''
Top level interface into Riffle library. The crust sits atop the mantle, which sits atop the core. 
The mantle is imported here as pymantle. It doesn't add functionality on top of the core, just
translates between coreRiffle and python as needed. 

Fixing the osx python version: 
    https://github.com/Homebrew/homebrew/blob/master/share/doc/homebrew/Common-Issues.md#python-segmentation-fault-11-on-import-some_python_module
'''

import os
import json

from greenlet import greenlet

import pymantle
from riffle import cumin, utils


class Domain(object):

    def __init__(self, name, superdomain=None, sibling=None):
        self.name = name

        if superdomain is not None:
            self.mantleDomain = superdomain.mantleDomain.Subdomain(name)
            self.app = superdomain.app
        elif sibling is not None:
            self.mantleDomain = sibling.mantleDomain.LinkDomain(name)
            self.app = sibling.app
        else:
            self.mantleDomain = pymantle.NewDomain(name)
            self.app = App()

    def join(self):
        # TODO: convert the "control plane" to use deferreds using Twisted style callbacks
        cb, eb = utils.newID(2)
        self.app.control[cb] = self.onJoin
        self.mantleDomain.Join(cb, eb)

        self.app.mainGreenlet = greenlet(self.app.handle)
        self.app.mainGreenlet.switch(self.mantleDomain)

    def leave(self):
        # TODO: Emit a deferred here (?)
        self.mantleDomain.Leave()

    def onJoin(self):
        pymantle.Info("Default onJoin")

    def onLeave(self):
        pymantle.Info("Default onLeave")

    def subscribe(self, endpoint, handler):
        return self._setHandler(endpoint, handler, self.mantleDomain.Subscribe, False)

    def register(self, endpoint, handler):
        return self._setHandler(endpoint, handler, self.mantleDomain.Register, True)

    def publish(self, endpoint, *args):
        return self._invoke(endpoint, args, self.mantleDomain.Publish, False)

    def call(self, endpoint, *args):
        return self._invoke(endpoint, args, self.mantleDomain.Call, True)

    def _setHandler(self, endpoint, handler, coreFunction, doesReturn):
        '''
        Register or Subscribe. Invokes targetFunction for the given endpoint and handler.

        :param coreFunction: the intended core function, either Subscribe or Register
        :param doesReturn: True if this handler can return a value (is a registration)
        '''

        d, handlerId = Deferred(), utils.newID()
        self.app.deferreds[d.cb], self.app.deferreds[d.eb] = d, d
        self.app.handlers[handlerId] = handler, doesReturn

        coreFunction(endpoint, d.cb, d.eb, handlerId, cumin.reflect(handler))
        return d

    def _invoke(self, endpoint, args, coreFunction, doesReturn):
        '''
        Publish or Call. Invokes targetFunction for the given endpoint and handler.

        :param coreFunction: the intended core function, either Subscribe or Register
        :param doesReturn: True if this handler can receive results (is a call)
        '''

        d = Deferred()
        d.canReturn = doesReturn
        d.mantleDomain = self.mantleDomain
        self.app.deferreds[d.cb], self.app.deferreds[d.eb] = d, d

        coreFunction(endpoint, d.cb, d.eb, cumin.marshall(args))
        return d

    def link(self, name):
        """
        Access another domain.

        Returns a new Domain object with the given name, which should
        be fully-qualified (e.g. "xs.demo.user.app.Storage").
        """
        return Domain(name, sibling=self)


class Deferred(object):

    def __init__(self):
        self.cb, self.eb = utils.newID(2)
        self.green = None

        # Boy does it hurt me to put this here. canReturn needs to have access to the domain
        # for call returns
        self.mantleDomain = None

        # True if this deferred should inform the core of its types once set
        # Only true for calls
        self.canReturn = False

    def wait(self, *types):
        ''' Wait until the results of this invocation are resolved '''

        # if canReturn then this is a call. We need to retroactively inform the core of our types
        if self.canReturn:
            self.mantleDomain.CallExpects(self.cb, cumin.prepareSchema(types))

        # Get our current greenlet. If we're not running in a greenlet, someone screwed up bad
        self.green = greenlet.getcurrent()

        # Switch back to the parent greenlet: block until the parent has time to resolve us
        results = self.green.parent.switch(self)

        if isinstance(results, Exception):
            raise results

        r = cumin.unmarshall(results, types)

        # Actually, this cant happen. A return from a function should always come back as
        # a list. Unless you mean a return from... any other thing. In which case, bleh.
        if r is None:
            return r

        return r[0] if len(r) == 1 else r

    def setCallback(self, handler):
        ''' Traditional Twisted style callbacks '''
        pass


class App(object):

    def __init__(self):
        # self.handlers contains a tuple of (function, bool), where the bool is True if the
        # function can return (i.e. is a registration)
        self.deferreds, self.handlers, self.control = {}, {}, {}
        self.mainGreenlet = None

    def handle(self, domain):
        ''' Open connection with the core and begin handling callbacks '''

        while True:
            i, args = json.loads(domain.Receive())
            args = [] if args is None else args

            # When the channel that feeds the Receive function closes, 0 is returned
            # This is currently automagic and happened to work on accident, but consider
            # a more explicit close
            if i == 0:
                break

            if i in self.deferreds:
                d = self.deferreds[i]
                del self.deferreds[d.cb]
                del self.deferreds[d.eb]

                # Special Python case-- if this is an errback construct an excepction
                if i == d.eb:
                    args = utils.Error(*args)

                # Deferreds are always emitted by async methods. If the user called .wait()
                # then the deferred instantiates a greenlet as .green. Resume that greenlet.
                # If that greenlet emits another deferred the user has called .wait() again.
                if d.green is not None:
                    d = d.green.switch(args)
                    if d is not None:
                        self.deferreds[d.cb], self.deferreds[d.eb] = d, d

            elif i in self.handlers:
                handler, canReturn = self.handlers[i]

                if canReturn:
                    resultID = args.pop(0)

                # Consolidated handlers into one
                try:
                    ret = handler(*args)
                except Exception as error:
                    # TODO: Differentiate Riffle errors vs. other exceptions,
                    # maybe reraise other exceptions.
                    if canReturn:
                        domain.YieldError(resultID, "wamp.error.runtime_error", json.dumps([str(error)]))
                    else:
                        print "An exception occured: ", error
                else:
                    if canReturn:
                        if ret is None:
                            ret = []
                        elif isinstance(ret, tuple):
                            ret = list(ret)
                        else:
                            ret = [ret]

                        domain.Yield(resultID, cumin.marshall(ret))

            # Control messages. These should really just be deferreds, but not implemented yet
            if i in self.control:
                d = greenlet(self.control[i]).switch(*args)

                # If user code called .wait(), this deferred is emitted, waiting on the results of some operation
                if d is not None:
                    self.deferreds[d.cb], self.deferreds[d.eb] = d, d



'''
Random utilities
'''

import random

def newID(n=1):
    ''' Returns n random unsigned integers to act as Callback Ids '''
    return random.getrandbits(53) if n == 1 else tuple([random.getrandbits(53) for x in range(n)])


class Error(Exception):
    def __init__(self, *args):
        super(Error, self).__init__(*args)
        self.message = Error.messageFromArgs(args)

    def __str__(self):
        return self.message

    @classmethod
    def messageFromArgs(_class, args):
        """
        Find the human-readable message in a list of args.

        args must be a tuple or list of any length.
        """
        if len(args) == 0:
            return ""
        elif len(args) == 1:
            return str(args[0])
        else:
            first = str(args[0])
            if first.startswith("wamp.error"):
                return str(args[1])
            else:
                return first


class SyntaxError(Error):
    pass

class CuminError(Error):
    pass

class Unimplemented(Error):
    pass



def test_unmarshall():
    from riffle.cumin import unmarshall

    # types set to None just returns the arguments
    assert unmarshall([42], None) == (42, )

    assert unmarshall([1, 2], [int, int]) == (1, 2)
    assert unmarshall([1, "a"], [int, str]) == (1, "a")


def test_unmarshall_object():
    from riffle.cumin import unmarshall
    from riffle.model import ModelObject

    class Person(ModelObject):
        name = str

    dale = Person(name="Dale")
    dale_s = dale._serialize()

    lance = Person(name="Lance")
    lance_s = lance._serialize()

    # Person objects as arguments
    assert unmarshall([dale_s], [Person]) == (dale, )
    assert unmarshall([dale_s, lance_s], [Person, Person]) == (dale, lance)

    # List of Person objects as an argument
    assert unmarshall([[]], [[Person]]) == ([], )
    assert unmarshall([[dale_s]], [[Person]]) == ([dale], )
    assert unmarshall([[dale_s, lance_s]], [[Person]]) == ([dale, lance], )



'''
Base implementation of Model class
'''

import sys
import os

class ModelObject(object):
    __collection = None
    __storage = None

    @classmethod
    def reflect(klass):
        '''
        Build a schema from the types listed as class variables and return it.

        Can also store it as a class variable. 

        The format of the returned data is: 
            {'email': 'str', 'name': 'str'}
        '''

        attrs = {}

        for s in dir(klass):

            # Skip internal stuff
            if(s.startswith("_")):
                continue

            a = getattr(klass, s)

            # If they defined just a type, call it so it returns the default value
            if(isinstance(a, type)):
                a = a.__call__()

            # Skip functions
            elif(hasattr(a, "__call__")):
                continue

            attrs[s] = type(a).__name__

        # klass.__schema = attrs
        return attrs

    def __init__(self, **kwargs):
        self.__attrs = dict()
        self.__values = dict()

        # Scan everything in the class
        for s in dir(self):

            # Skip internal stuff
            if(s.startswith("_")):
                continue

            a = getattr(self, s)

            # If they defined just a type, call it so it returns the default value
            if(isinstance(a, type)):
                a = a.__call__()

            # Skip functions
            elif(hasattr(a, "__call__")):
                continue

            self.__attrs[s] = type(a)
            self.__values[s] = a

        # Update with values that were set as arguments.
        self.__values.update(kwargs)

        # Every object that is in the storage will have an _id field.
        self.__id = None

    def __getattribute__(self, name):
        if name.startswith("_") or name not in self.__values:
            return super(ModelObject, self).__getattribute__(name)
        else:
            return self.__values[name]

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super(ModelObject, self).__setattr__(name, value)
        else:
            if name not in self.__values:
                raise AttributeError("Class {} does not have an attribute {}"
                        .format(self.__class__.__name__, name))
            elif not isinstance(value, self.__attrs[name]):
                raise TypeError("For attribute {} expected type {}, received {}"
                        .format(name, self.__attrs[name].__name__,
                            type(value).__name__))
            else:
                self.__values[name] = value

    def __eq__(self, other):
        if isinstance(other, ModelObject):
            return other.__values == self.__values
        elif isinstance(other, dict):
            return other == self.__values
        else:
            return False

    def __repr__(self):
        return str(self.__class__.__name__) + repr(self.__values)

    @classmethod
    def _deserialize(klass, json):
        # TODO: apply deserialze recursively to nested models and collections of models
        return klass(**json)

    def _serialize(self):
        # TODO: apply serialize recursively to nested models
        # TODO: check values, make sure they're serializable, else throw an exception
        return self.__values

    #
    # Methods for model object persistence.
    #

    @classmethod
    def bind(klass, owner, collection=None, appliance=None):
        """
        Bind this model to a storage domain for persistence.

        If collection is not specified, the default behavior is to use the
        class name as the collection.

        owner: instance of a riffle.Domain
        collection: name of collection
        appliance: fully-qualified domain of Storage appliance if it differs
                   from the 'owner' object
        """
        if klass.__storage is not None:
            raise Exception("Model {} already bound".
                    format(klass.__name__))
        if collection is None:
            collection = klass.__name__
        klass.__collection = collection
        if appliance:
            klass.__storage = owner.link(appliance)
        else:
            klass.__storage = owner

    @classmethod
    def assertBound(klass):
        if klass.__storage is None:
            raise Exception("{model} is not bound; "
                    "you should call {model}.bind first"
                    .format(model=klass.__name__))

    def save(self):
        """
        Write the object back to storage.
        """
        self.assertBound()
        if self.__id is None:
            d = self.__storage.call("collection/insert_one",
                    self.__collection,
                    self._serialize())
            # TODO: d.wait({'$atleast': {'inserted_id': str}})
            result = d.wait(dict)
            self.__id = result.get('inserted_id', None)
            print("id: {}".format(self.__id))
        else:
            self.__storage.call("collection/replace_one",
                    self.__collection,
                    {'_id': self.__id},
                    self._serialize(),
                    True).wait()

    @classmethod
    def find(klass, query):
        """
        Find objects that match the query.

        Currently, supports equality checking only.

        Example:
        Users.find({name="Dale"})
        """
        klass.assertBound()
        d = klass.__storage.call("collection/find",
                klass.__collection, query)
        return d.wait([klass])

    @classmethod
    def find_one(klass, query):
        """
        Find a single object that matches the query.

        Currently, supports equality checking only.

        Example:
        Users.find({name="Dale"})
        """
        klass.assertBound()
        d = klass.__storage.call("collection/find_one",
                klass.__collection, query)
        return d.wait(klass)


# Model is deprecated.
#
# We should prefer ModelObject or some variation as a class name so that "is a"
# relationships make sense for instances of the class.
#
# Example: suppose Person is a subclass of ModelObject, and dale is an instance
# Person.  Logically, dale is a Person, dale is a ModelObject (an object that
# conforms to the Person model), but dale is not a Model.
Model = ModelObject



from riffle.model import ModelObject

def test_recursive_prepareSchema():
    from riffle.cumin import _prepareSchema

    assert _prepareSchema(None) is None

    assert _prepareSchema(int) == "int"
    assert _prepareSchema(float) == "float"
    assert _prepareSchema(bool) == "bool"
    assert _prepareSchema(str) == "str"
    assert _prepareSchema(list) == "list"
    assert _prepareSchema(dict) == "dict"

    assert _prepareSchema([str]) == ['str']

    assert _prepareSchema({'test': int}) == {'test': 'int'}

    class Person(ModelObject):
        name = str
    assert _prepareSchema(Person) == {'name': 'str'}



'''
Type reflection and conversion.
'''

import inspect
import json

from riffle import utils, model

def want(*types):
    def real_decorator(function):
        def wrapper(*args, **kwargs):
            # return the types this call expects as a list if asked
            if '_riffle_reflect' in kwargs:
                return list(types)

            # Test if function looks like a method and set aside the first
            # argument.  At first glance, inspect.ismethod should tell us this,
            # but it doesn't, so for now we just look for a "self" argument.
            pre_args = tuple()
            argspec = inspect.getargspec(function).args
            if argspec and argspec[0] == "self":
                pre_args = args[0:1]
                args = args[1:]

            final_args = pre_args + unmarshall(args, types)

            return function(*final_args)
        return wrapper
    return real_decorator

def reflect(handler):
    '''
    Reflectes the expected types of the receiver and returns them as a json serialized list.

    If no @want is specified, None is returned. This allows any arguments.
    '''
    try:
        types = handler(_riffle_reflect=True)
    except TypeError, e:
        # If the method does not accept a _riffle_reflect, its not a wrapped function and 
        # thus accepts any argument
        # TODO: HOWEVER-- a **kwargs function will fail here. Have to catch **kwargs functions
        # and immediately err on them! Receives can only accept *args and may embed as elements
        return json.dumps([None])

    return prepareSchema(types)


def _prepareSchema(t):
    if t is None:
        return None

    elif t in [int, float, bool, str, list, dict]:
        return t.__name__

    elif isinstance(t, list):
        if len(t) > 1:
            raise utils.SyntaxError("Lists can only have one internal type.")
        return [_prepareSchema(t[0])]

    elif isinstance(t, dict):
        tmp = dict()
        for k, v in t.iteritems():
            if not isinstance(k, basestring):
                raise utils.SyntaxError("Dictionary keys must be strings.")
            tmp[k] = _prepareSchema(v)
        return tmp

    elif issubclass(t, model.ModelObject):
        return t.reflect()

    else:
        raise utils.SyntaxError("Type {} is not natively serializible!"
                .format(str(t)))


def prepareSchema(types):
    '''
    Prepares a list of types for consumption by the core. Returns json.
    '''
    if types is None:
        return json.dumps([None])
    else:
        typeList = [_prepareSchema(t) for t in types]
        return json.dumps(typeList)

def marshall(args):
    ''' Ask models to serialize themselves, if present. Returns JSON. '''
    return json.dumps([x._serialize() if isinstance(x, model.ModelObject) else x for x in args])

def unmarshall(args, types):
    '''
    Prepares arguments from the core for user level code.

    Objects and exceptions are recreated from dictionaries if appropriate.
    '''

    l = _unmarshall(args, types)
    return tuple(l)

def _unmarshall(args, types):
    # If types is None, allow all arguments
    if not types:
        return args

    l = list()
    for x, y in zip(args, types):
        if isinstance(y, list):
            if len(y) == 1 and len(x) > 1:
                # e.g. x = [dale, mickey, lance] and y = [Person]
                l.append(_unmarshall(x, y * len(x)))
            else:
                l.append(_unmarshall(x, y))

        elif isinstance(y, type) and issubclass(y, model.ModelObject):
            l.append(y._deserialize(x))

        else:
            l.append(x)

    return l




from crust import Domain
from model import ModelObject
from cumin import want
from utils import Error, CuminError

from pymantle import SetLogLevelErr, SetLogLevelWarn, SetLogLevelInfo, SetLogLevelDebug, SetLogLevelApp, SetLogLevelOff
from pymantle import SetFabricDev, SetFabricSandbox, SetFabricProduction, SetFabricLocal, SetFabric
from pymantle import Application as log



from nose.tools import assert_raises

from riffle.model import ModelObject

def test_subclass():
    class Person(ModelObject):
        name = str
        age = int
        cool = False

    dale = Person(name="Dale")
    dale.cool = True

    # Value set through constructor
    assert dale.name == "Dale"

    # Value not set (default for int type)
    assert dale.age == 0

    # Value set through setattr
    assert dale.cool

    with assert_raises(AttributeError) as error:
        dale.missing = 5

    with assert_raises(TypeError) as error:
        dale.age = "bad"



from setuptools import setup, Extension

setup(
    name="pyRiffle",
    version="0.2.0",
    description="Riffle client libraries for interacting over a fabric",
    author="Exis",
    url="http://www.exis.io",
    license="MIT",

    packages=["riffle"],
    include_package_data=True,

    install_requires=[
        'greenlet'
    ],

    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)
